/*
 * $Source$
 * $Author$
 * $Revision$
 * 
 * relies heavily on STL. Could probably be made to run faster
 * when using std::hash_map instead of std::map, but that will
 * cause it to be less compatible. - Bram
 *
 * (c)2000 by Bram Stolk (b.stolk at host chello.nl)
 */

#include <cstdio>
#include <cstring>
#include <cstdlib>

#include <iostream>

#include "idep.h"


// helper func. replaces GNU basename()
static const char *strip_dir(const char *filename)
{
  int len = std::strlen(filename);
  if (!len) return filename;
  const char *p = filename + len - 1;
  while (*p && p>filename && *p != '/')
    p--;
  if (*p=='/') p++;
  return p;
}

// helper func
static char *strip_nl(char *line)
{
  int len = std::strlen(line);
  if (!len) return line;
  if (line[len-1]=='\n')
    line[len-1]=0;
  return line;
}

static char *strip_trailers(char *line, const char * const trailers)
{
  int len = std::strlen(line);
  if (!len)
  return line;
  char c = line[len-1];
  if (std::strchr(trailers, c))
  {
    line[len-1]=0;
    return strip_trailers(line, trailers);
  }
  return line;
}


HeaderInfo::HeaderInfo(const std::string &fullnm) :
  fullname(fullnm)
{
  shortname = strip_dir(fullnm.c_str());
}


void DepChecker::BuildDb(void)
{
  std::string cmd = 
    "find " + dirname + " -name \\*" + extention + " -print";
 
  std::FILE *f = ::popen(cmd.c_str(), "r");
  if (!f)
  {
    std::cerr << "cmd " << cmd << " failed.\n";
    std::exit(1);
  }

  char line[512];
  while (std::fgets(line, sizeof(line), f))
  {
    char *fulln = strip_nl(line);
    HeaderInfo hi(fulln);
    mapping[hi.shortname] = hi;
  }
  ::pclose(f);
}


void DepChecker::SearchForDeps(void)
{
  mapping_t::iterator ip;
  for (ip=mapping.begin(); ip!=mapping.end(); ++ip)
  {
    const std::string &fn = ip->second.fullname;
    std::string cmd = 
      "grep \"#include\" " + fn;
    std::FILE *f = ::popen(cmd.c_str(), "r");

    if (!f)
    {
      std::cerr << "cmd " << cmd << " failed.\n";
      std::exit(1);
    }
    char line[512];
    while (std::fgets(line, sizeof(line), f))
    {
      char s[256];
      bool skip=false;
      if (std::sscanf(line, " #include <%s> ", s) != 1)
      {
        if (std::sscanf(line, " #include \"%s\" ", s) != 1) 
        {
//          std::cerr << "Cannot interpret include:\n" << line;
          skip=true;
        }
      }
      if (!skip)
      {
        if (s[strlen(s)-1] == '>' || s[strlen(s)-1] == '"')
          s[strlen(s)-1] = 0;
        std::string dep = s;
        if (mapping.count(s) == 1)
          ip->second.deps.insert(ip->second.deps.end(), s);
      }
    }
    ::pclose(f);
  }
}


void DepChecker::SearchForDefs(void)
{
  mapping_t::iterator ip;
  for (ip=mapping.begin(); ip!=mapping.end(); ++ip)
  {
    HeaderInfo &hi = ip->second;
    const std::string &fn = hi.fullname;
    std::string cmd = 
      "grep \"class\" " + fn + " | grep -v \\;";

    std::FILE *f = ::popen(cmd.c_str(), "r");

    if (!f)
    {
      std::cerr << "cmd " << cmd << " failed.\n";
      std::exit(1);
    }
    char line[512];
    while (fgets(line, sizeof(line), f))
    {
      char name[256];
      if (std::sscanf(line, " class %s ", name) == 1)
      {
        const char *stripped_name = strip_trailers(name, ":");
        hi.defs.insert(hi.defs.begin(), stripped_name);
      }
    }
    ::pclose(f);
  }
}


bool DepChecker::DirectUse(std::string user, std::string usedclass)
{
  const std::string fulln = mapping[user].fullname;
  const std::string cmd = 
    "grep " + usedclass + " " + fulln + 
    " | grep -v \"" + usedclass + "[ ]*[\\*&]\" " +
    " | grep -v \"#include\"";
 
  std::FILE *f = ::popen(cmd.c_str(), "r");

  if (!f)
  {
    std::cerr << "cmd " << cmd << " failed.\n";
    std::exit(1);
  }
  char line[512];
  int cnt=0;
  while (std::fgets(line, sizeof(line), f))
    cnt++;

  ::pclose(f);

  return cnt>0;
}


void DepChecker::SearchForPrunables(void)
{
  // Traverse all header files
  mapping_t::iterator ip1;
  for (ip1=mapping.begin(); ip1!=mapping.end(); ++ip1)
  {
    std::string user = ip1->first;
    HeaderInfo &hi1 = ip1->second;
    // Traverse all included files of this header file
    stringset_t::iterator ip2;
    for (ip2=hi1.deps.begin(); ip2!=hi1.deps.end(); ++ip2)
    {
      std::string usee = *ip2;
      // Traverse all class defs of include of header file
      HeaderInfo &hi2 = mapping[usee];
      stringset_t::iterator ip3;
      bool required = false;
      // If the header file does not define classes, it probably
      // contains contant defs only, and is likely to be required.
      if (hi2.defs.size() == 0)
        required = true;
      for (ip3=hi2.defs.begin(); ip3!=hi2.defs.end() && !required; ++ip3)
      {
        std::string usedclass = *ip3;
        bool needed = DirectUse(user, usedclass);
        if (needed)
          required = true;
      }
      if (!required)
      {
#if 0
        std::cout << "file " << user << 
          " does not require the inclusion of " <<
          usee << " (indirect refs only!)" << std::endl;
#endif
        hi1.prunables.insert(hi1.prunables.end(), usee);
      }
    }
  }
}



void DepChecker::PrintNode
(
  const std::string name, 
  const std::string indent, 
  const std::string prefix,
  std::set<std::string> visited,
  std::set<std::string> &printed
)
{
  std::cout << indent << prefix << name << " [ ";
  const HeaderInfo &hi = mapping[name];

  for (stringset_t::const_iterator i=hi.defs.begin(); i!=hi.defs.end(); ++i)
  {
    std::cout << *i << " ";
  }

  std::cout << "]" << std::endl;

  if (visited.count(name))
  {
    std::cout << indent << "CYCLIC DEPENDENCY" << std::endl;
    return;
  }

  if (printed.count(name))
  {
    if (hi.deps.size())
      std::cout << indent << "  ..." << std::endl;
    // Stop recursion. We've printed this branch before!
    return;
  }

  printed.insert(printed.begin(), name);

  visited.insert(visited.begin(), name);
  const stringset_t &deps = hi.deps;
  stringset_t::const_iterator ip;
  for (ip=deps.begin(); ip!=deps.end(); ++ip)
  {
    std::string prefix="";
    if (hi.prunables.count(*ip))
      prefix="PRUNABLE: ";
    else
      prefix="";
    PrintNode(*ip, indent+"  ", prefix, visited, printed);
  }
}



void DepChecker::Print(void)
{
  mapping_t::const_iterator ip;
  for (ip=mapping.begin(); ip!=mapping.end(); ++ip)
  {
    const std::string &name = ip->first;
    if (IsRootNode(name))
    {
      std::cout << std::endl;
      std::set<std::string> emptyset_v;
      std::set<std::string> emptyset_p;
      const std::string &name = ip->first;
      PrintNode(name, "", "ROOT:", emptyset_v, emptyset_p);
    }
  }
}


bool DepChecker::IsRootNode(std::string name) const
{
  for (mapping_t::const_iterator ip=mapping.begin(); ip!=mapping.end(); ++ip)
  {
    if (ip->second.deps.count(name))
      return false;
  }
  return true;
}


int main(int argc, char *argv[]) 
{
  if (argc!=3) 
  {
    std::cerr << "Usage: " << argv[0] << " dirname extention" << std::endl;
    return 1;
  }
 
  DepChecker dc(argv[1], argv[2]);
  std::cout << ">> Building database" << std::endl;
  dc.BuildDb();
  std::cout << ">> Searching for dependencies" << std::endl;
  dc.SearchForDeps();
  std::cout << ">> Searching for class definitions" << std::endl;
  dc.SearchForDefs();
  std::cout << ">> Searching for prunable includes" << std::endl;
  dc.SearchForPrunables();

  std::cout << ">> Dependency tree:" << std::endl;
  dc.Print();
}

